﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Praktikum_3_Loesung
{
    class EAuto
    {
        string kennzeichen;
        public string Kennzeichen { get { return kennzeichen; } }

        double maximaleReichweite; // bei vollem Akku, in Kilometern
        public double MaximaleReichweite { get { return maximaleReichweite; } }

        double ladezeit; // Zeit zum vollständigen Laden bei leerem Akku, in Stunden
        public double Ladezeit { get { return ladezeit; } }

        public EAuto(string kennzeichen, double maximaleReichweite, double ladezeit)
        {
            this.kennzeichen = kennzeichen;
            this.maximaleReichweite = maximaleReichweite;
            this.ladezeit = ladezeit;
        }

        public override string ToString()
        {
            return kennzeichen;
        }
    }

    class Buchung
    {
        Buchungszeitraum zeitraum;
        public Buchungszeitraum Zeitraum { get { return zeitraum; } }

        EAuto fahrzeug;
        public EAuto Fahrzeug { get { return fahrzeug; } }

        double strecke;
        public double Strecke { get { return strecke; } }

        string wer;
        public string Wer { get { return wer; } }

        public Buchung(string wer, EAuto fahrzeug, DateTime von, DateTime bis, double strecke)
        {
            this.wer = wer;
            this.fahrzeug = fahrzeug;
            zeitraum = new Buchungszeitraum(von, bis);
            this.strecke = strecke;
        }

        public override string ToString()
        {
            return Wer + " " + Fahrzeug + " " + Zeitraum + " " + Strecke + " km";
        }
    }

    class Buchungszeitraum
    {
        DateTime von;
        public DateTime Von { get { return von; } }

        DateTime bis;
        public DateTime Bis { get { return bis; } }

        public Buchungszeitraum(DateTime von, DateTime bis)
        {
            if (von > bis)
            {
                throw new ApplicationException("falsche zeitliche Reihenfolge");
            }

            this.von = von;
            this.bis = bis;
        }

        // Es ist erlaubt, dass die Zeiträume genau aneinander angrenzen.
        public bool ÜberschneidetSichNichtMit(Buchungszeitraum z)
        {
            return this.bis <= z.von || z.bis <= this.von;
        }

        public bool ÜberschneidetSichMit(Buchungszeitraum z)
        {
            return !ÜberschneidetSichNichtMit(z);
        }

        public double ZeitDanachBis(Buchungszeitraum z) // in Stunden
        {
            return (z.Von - this.Bis).TotalHours;
        }

        public override string ToString()
        {
            return von + " bis " + bis;
        }
    }

    class Verwaltung
    {
        List<Buchung> buchungen = new List<Buchung>();

        public bool VersucheBuchung(string wer, EAuto fahrzeug, DateTime von, DateTime bis, double strecke)
        {
            List<Buchung> buchungenFürDiesesFahrzeug = new List<Buchung>();
            foreach (Buchung b in buchungen)
            {
                if (b.Fahrzeug == fahrzeug)
                {
                    buchungenFürDiesesFahrzeug.Add(b);
                }
            }

            Buchung buchung = new Buchung(wer, fahrzeug, von, bis, strecke);

            foreach (Buchung b in buchungenFürDiesesFahrzeug)
            {
                if (b.Zeitraum.ÜberschneidetSichMit(buchung.Zeitraum))
                {
                    return false;
                }
            }

            buchungenFürDiesesFahrzeug.Add(buchung);

            // Der folgende Code prüft,
            // ob die Ladezeiten zwischen den Buchungen reichen,
            // um genügend Reichweite für die kommenden Fahrten zu haben.

            List<Buchung> buchungenZeitlichSortiert
                = buchungenFürDiesesFahrzeug.OrderBy(b => b.Zeitraum.Von).ToList();

            // Der Akku ist zu Beginn voll geladen.
            double reichweite = fahrzeug.MaximaleReichweite;

            for (int i = 0; i < buchungenZeitlichSortiert.Count; i++)
            {
                reichweite -= buchungenZeitlichSortiert[i].Strecke;
                if (reichweite < 0.0)
                {
                    return false;
                }

                // Die Zeit bis zur nächsten Buchung zum Laden benutzen.
                if (i < buchungenZeitlichSortiert.Count - 1)
                {
                    double zeitZumLaden = buchungenZeitlichSortiert[i].Zeitraum.ZeitDanachBis(buchungenZeitlichSortiert[i + 1].Zeitraum);
                    double erhöhungDerReichweite = fahrzeug.MaximaleReichweite * zeitZumLaden / fahrzeug.Ladezeit;
                    reichweite += erhöhungDerReichweite;
                    // Man kann nicht mehr als voll laden.
                    if (reichweite >= fahrzeug.MaximaleReichweite)
                    {
                        reichweite = fahrzeug.MaximaleReichweite;
                    }
                }
            }

            buchungen.Add(buchung);
            return true;
        }

        public string ListeBuchungen()
        {
            StringBuilder sb = new StringBuilder(); // effizienter als string, aber string ginge auch.
            foreach (var buchung in buchungen)
            {
                sb.AppendLine(buchung.ToString());
            }
            return sb.ToString();
        } 
    }
}
